/*
 * $Id: gp.c,v 1.10 2001/09/13 14:21:37 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */

/**
 * Galactica Proxy main thread.
 *
 * @version $Revision: 1.10 $
 */
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#if defined(__APPLE__) && defined(__MACH__)
#include <OpenGL/gl.h>	// Header File For The OpenGL32 Library
#include <OpenGL/glu.h>	// Header File For The GLu32 Library
#else
#include <GL/gl.h>	// Header File For The OpenGL32 Library
#include <GL/glu.h>	// Header File For The GLu32 Library
#endif
#include "SDL.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "gp.h"
#include "screen.h"

#define FRAGS_SIZE 16

static Hashtable hosts;
static Frag frags[FRAGS_SIZE];
static int fragIdx = 0;

static Page *currentPage = NULL, *prvCp = NULL;
static int cpHostIdx = -1, cpPageIdx = -1;

// Save/Load history tree to/from file.
#define HISTORY_FILE "history.cfg"
#define VER_KEY "GP012"

static int checkVerKey(FILE *fp) {
  char *verKey = VER_KEY;
  int i;
  for ( i=0 ; i<strlen(verKey) ; i++ ) {
    if ( getw(fp) != verKey[i] ) return 0;
  }
  return 1;
}

static void writeVerKey(FILE *fp) {
  char *verKey = VER_KEY;
  int i;
  for ( i=0 ; i<strlen(verKey) ; i++ ) {
    putw(verKey[i], fp);
  }
}

static int loadHistory() {
  FILE *fp = fopen(HISTORY_FILE, "rb");
  if ( fp == NULL ) {
    fprintf(stderr, "Unable to load history file.\n");
    return 0;
  }
  if ( !checkVerKey(fp) ) {
    fprintf(stderr, "History file version is changed.\n");
    return 0;
  }
  load(&hosts, fp);
  startEnumeration(&hosts);
  while ( hasMoreElements(&hosts) ) {
    Host* host = nextElement(&hosts);
    Hashtable *pages = &(host->pages);
    host->start = 2; host->txWidth = 0; host->isActive = 1;
    load(pages, fp);
    startEnumeration(pages);
    while ( hasMoreElements(pages) ) {
      Page* page = nextElement(pages);
      Hashtable *blinks = &(page->bLinks);
      page->start = 2; page->isActive = 1;
      if ( page->drawnTitle == 1 ) {
	page->drawnTitle = 0;
      }
      load(blinks, fp);
    }
  }
  cpHostIdx = getw(fp);
  cpPageIdx = getw(fp);
  if ( cpHostIdx >= 0 && cpPageIdx >= 0 ) {
    Host *ch = checkByIndex(&hosts, cpHostIdx);
    currentPage = checkByIndex(&(ch->pages), cpPageIdx);
  }
  fclose(fp);
  return 1;
}

static void saveHistory(SDL_mutex *lock) {
  FILE *fp = fopen(HISTORY_FILE, "wb");
  if ( fp == NULL ) {
    fprintf(stderr, "Unable to save history file.\n");
    return;
  }
  SDL_mutexP(lock);
  writeVerKey(fp);
  save(&hosts, fp);
  startEnumeration(&hosts);
  while ( hasMoreElements(&hosts) ) {
    Hashtable *pages = &(((Host*)nextElement(&hosts))->pages);
    save(pages, fp);
    startEnumeration(pages);
    while ( hasMoreElements(pages) ) {
      Hashtable *blinks = &(((Page*)nextElement(pages))->bLinks);
      save(blinks, fp);
    }
  }
  putw(cpHostIdx, fp);
  putw(cpPageIdx, fp);
  SDL_mutexV(lock);
  fclose(fp);
}

static void init() {
  int i, j;
  Host **hostElms;
  Page **pageElms;
  for ( i=0 ; i<FRAGS_SIZE ; i++ ) {
    generateTexture(&(frags[i].texture));
    frags[i].cnt = 0;
  }
  if ( !loadHistory() ) {
    initHash(&hosts, sizeof(Host), 1);
  }
  srand(SDL_GetTicks());
}

// Make string length < n.
static void setStrLgt(char *str, int n) {
  int lgt = strlen(str);
  if ( lgt >= n ) {
    str[n-1] = '\0';
  }
}

// Trim string.
static char* trimStr(char *str) {
  int i;
  for ( i=0 ; i<strlen(str) ; i++ ) {
    if ( str[i] != ' ' && str[i] != '\n' && str[i] != '\r' ) break;
  }
  return &(str[i]);
}

// Add not html files twinkle stars.
static void addFrag(char *hostname, char *pathname) {
  char frNm[HOST_NAME_LENGTH + PATH_NAME_LENGTH];
  char *fnp;
  strcpy(frNm, hostname);
  strcat(frNm, pathname);
  fnp = strrchr(frNm, '/');
  if ( fnp == NULL ) {
    fnp = frNm;
  } else {
    fnp++;
  }
  setStrLgt(fnp, FRAG_NAME_LENGTH);
  strcpy(frags[fragIdx].name, fnp);
  frags[fragIdx].start = 1;
  frags[fragIdx].cnt = 1;
  fragIdx++; fragIdx &= (FRAGS_SIZE-1);
}

// The host/page is over written.

// Textures should be deleted.
static void deleteTexturesPage(Page *page) {
  deleteTexture(&(page->texture));
}

static void deleteTexturesHost(Host *host) {
  Hashtable *pages = &(host->pages);
  startEnumeration(pages);
  while ( hasMoreElements(pages) ) {
    Page *page = nextElement(pages);
    deleteTexturesPage(page);
  }
  deleteTexture(&(host->texture));
}

// Links to this host/page is not valid.
static void deprecateLinkPage(int hostkey, int pagekey) {
  startEnumeration(&hosts);
  while ( hasMoreElements(&hosts) ) {
    Host *host = nextElement(&hosts);
    Hashtable *pages = &(host->pages);
    startEnumeration(pages);
    while ( hasMoreElements(pages) ) {
      Page *page = nextElement(pages);
      Hashtable *blinks = &(page->bLinks);
      startEnumeration(blinks);
      while ( hasMoreElements(blinks) ) {
	Link *blink = nextElement(blinks);
	if ( hostkey == blink->hostIdx ) {
	  if ( pagekey == -1 || pagekey == blink->pageIdx ) {
	    deleteElement(blinks);
	  }
	}
      }
    }
  }
}

static void deprecateLinkHost(int hostkey) {
  deprecateLinkPage(hostkey, -1);
}

#define VALID_PAGE_SIZE 1024

// Check a command from the user agent.
void hookProxyConn(char *hostname, char *pathname, char *gpc, char *referer, 
		   int isHtml, char *title, int bodySize,
		   SDL_mutex *lock) {
  Host *host;
  Page *page;
  char* rfIdx;
  int valid, new;
  int isPage = 0;

  if ( strstr(gpc, "GET") == NULL ) return;

  if ( hostname != NULL ) {
    setStrLgt(hostname, HOST_NAME_LENGTH);
  } else {
    hostname = "?";
  }
  if ( pathname != NULL ) {
    setStrLgt(pathname, PATH_NAME_LENGTH);
  } else {
    pathname = "";
  }

  if ( bodySize == 0 ) {
    // BodySize == 0, perhaps page content is not changed.
    //printf("not changed.\n");
    host = check(&hosts, hostname, &valid);
    if ( valid ) {
      page = check(&(host->pages), pathname, &valid);
      if ( valid ) {
	prvCp = currentPage;
	currentPage = page;
	cpHostIdx = hashFunc(hostname);
	cpPageIdx = hashFunc(pathname);
	isPage = 1;
      }
    }
  } else if ( !isHtml || bodySize < VALID_PAGE_SIZE ) {
    // Not text/html page.
    SDL_mutexP(lock);
    addFrag(hostname, pathname);
    SDL_mutexV(lock);
  } else {
    // Valid page.
    int hostIsNew = 1;
    //printf("host: %s\n", hostname);
    SDL_mutexP(lock);
    isPage = 1;
    host = get(&hosts, hostname, &valid, &new);
    if ( !valid ) {
      if ( !new ) {
	deleteTexturesHost(host);
	deprecateLinkHost(hashFunc(hostname));
	hostIsNew = 0;
      }
      initHash(&(host->pages), sizeof(Page), new);
      host->start = 1;
      host->isActive = 1;
      host->txWidth = 0;
      strcpy(host->name, hostname);
    }
    page = get(&(host->pages), pathname, &valid, &new);
    //printf("path: %s\n", pathname);
    if ( !valid ) {
      if ( hostIsNew && !new ) {
	deleteTexturesPage(page);
	deprecateLinkPage(hashFunc(hostname), hashFunc(pathname));
      }
      initHash(&(page->bLinks), sizeof(Link), new);
      page->host = host;
      if ( bodySize < 8192 ) {
	page->r = bodySize/32; page->g = page->b = 64;
      } else if ( bodySize < 16384 ) {
	page->r = 255; page->g = (bodySize-8192)/32; page->b = 64;
      } else {
	page->r = page->g = 255-(bodySize/400);
	if ( page->r < 128 ) page->r = page->g = 128;
	page->b = bodySize/200; if ( page->b >= 256 ) page->b = 255;
      }
      host->isActive = 1;
      page->isActive = 1;
      page->start = 1;
      page->title[0] = '\0';
      page->drawnTitle = 2;
      if ( title != NULL ) {
	title = trimStr(title);
	if ( strlen(title) > 0 ) {
	  //printf("title: %s\n", title);
          setStrLgt(title, TITLE_LENGTH);
	  strcpy(page->title, title);
	  if ( currentPage == NULL || 
	       strcmp(currentPage->title, title) != 0 ) {
	    page->drawnTitle = 0;
	  }
	}
      }
      page->txWidth = 0;
      strcpy(page->path, pathname);
    }
    prvCp = currentPage;
    currentPage = page;
    cpHostIdx = hashFunc(hostname);
    cpPageIdx = hashFunc(pathname);
    SDL_mutexV(lock);
  }

  if ( isPage ) {
    if ( referer != NULL &&
	 (rfIdx = strstr(referer, "Referer: http://")) != NULL ) {
      char rfHst[HOST_NAME_LENGTH], *rfPath, 
	rfKey[HOST_NAME_LENGTH+PATH_NAME_LENGTH];
      //printf("referer: %s\n", referer);
      rfIdx += 16;
      rfPath = strchr(rfIdx, '/');
      if ( rfPath == NULL ) {
	setStrLgt(rfIdx, HOST_NAME_LENGTH);
	strcpy(rfHst, rfIdx); 
	rfPath = "";
      } else {
	int rhLgt = rfPath - rfIdx;
	if ( rhLgt >= HOST_NAME_LENGTH ) rhLgt = HOST_NAME_LENGTH-1;
	strncpy(rfHst, rfIdx, rhLgt);
	rfHst[rhLgt] = '\0';
	setStrLgt(rfPath, PATH_NAME_LENGTH);
      }
      SDL_mutexP(lock);
      host = check(&hosts, rfHst, &valid);
      if ( valid ) {
	page = check(&(host->pages), rfPath, &valid);
	if ( valid ) {
	  Link *blk;
	  //printf("add link.\n");
	  strcpy(rfKey, rfHst);
	  strcat(rfKey, rfPath);
	  blk = get(&(currentPage->bLinks), rfKey, &valid, &new);
	  if ( !valid ) {
	    blk->hostIdx = hashFunc(rfHst);
	    blk->pageIdx = hashFunc(rfPath);
	  }
	}
      }
      SDL_mutexV(lock);
    }
  }
}

// Set the initial position and movment of the star.
#define HOST_INIT_MV 0.11

static float makeHostMvRand() {
  if ( rand() > RAND_MAX/2 ) {
    return ((float)rand()/RAND_MAX)*HOST_INIT_MV/2 + HOST_INIT_MV/2;
  } else {
    return -((float)rand()/RAND_MAX)*HOST_INIT_MV/2 - HOST_INIT_MV/2;
  }
}

static void setHostInitPos(Host *host, Page *cp) {
  if ( cp != NULL ) {
    host->x = cp->x; host->y = cp->y; host->z = cp->z;
  } else {
    host->x = host->y = host->z = 0;
  }
  host->mx = makeHostMvRand();
  host->my = makeHostMvRand();
  host->mz = makeHostMvRand();
}

#define PAGE_INIT_MV 0.04

static void setPageInitPos(Page *page, Page *cp) {
  if ( cp != NULL ) {
    page->x = cp->x; page->y = cp->y; page->z = cp->z;
  } else {
    page->x = page->y = page->z = 0;
  }
  page->mx = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
  page->my = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
  page->mz = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
}

#define FRAG_INIT_MV 0.03

static void setFragInitPos(Frag *frag, Page *cp) {
  if ( cp != NULL ) {
    frag->x = cp->x; frag->y = cp->y; frag->z = cp->z;
  } else {
    frag->x = frag->y = frag->z = 0;
  }
  frag->mx = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
  frag->my = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
  frag->mz = ((float)(rand()-RAND_MAX/2)/(RAND_MAX/2))*PAGE_INIT_MV;
}

#define SLOWDOWN 0.99f
#define COLORDOWN 0.98f
#define HOST_RANGE 7
#define HOST_PULL 500
#define MOVE_IS_ACTIVE 0.001f

static void move(SDL_mutex *lock) {
  int i;
  startEnumeration(&hosts);
  while ( hasMoreElements(&hosts) ) {
    Host *host = nextElement(&hosts);
    Hashtable *pages;
    if ( host->isActive ) {
      if ( host->start != 0 ) {
	SDL_mutexP(lock);
	if ( host->txWidth == 0 ) {
	  makeTextTexture(host->name, &(host->texture),
			  &(host->txWidth), &(host->txHeight));
	}
	if ( host->start == 1 ) {
	  setHostInitPos(host, prvCp);
	}
	host->start = 0;
	SDL_mutexV(lock);
      }
      host->x += host->mx; host->y += host->my; host->z += host->mz;
      host->mx *= SLOWDOWN; host->my *= SLOWDOWN; host->mz *= SLOWDOWN;
      if ( fabs(host->mx) < MOVE_IS_ACTIVE && 
	   fabs(host->my) < MOVE_IS_ACTIVE &&
	   fabs(host->mz) < MOVE_IS_ACTIVE ) {
	host->isActive = 0;
      }
      pages = &(host->pages);
      startEnumeration(pages);
      while ( hasMoreElements(pages) ) {
	Page *page = nextElement(pages);
	float dst;
	if ( host->isActive || page->isActive ) {
	  if ( page->start != 0 ) {
	    SDL_mutexP(lock);
	    if ( page->drawnTitle == 0 ) {
	      page->drawnTitle = 1;
	      makeTextTexture(page->title, &(page->texture),
			      &(page->txWidth), &(page->txHeight));
	    }
	    if ( page->start == 1 ) {
	      setPageInitPos(page, prvCp);
	    }
	    page->start = 0;
	    SDL_mutexV(lock);
	  }
	  dst = abs(page->x - host->x) 
	    + abs(page->y - host->y) + abs(page->z - host->z);
	  if ( dst > HOST_RANGE ) {
	    page->mx += (host->x - page->x)/HOST_PULL;
	    page->my += (host->y - page->y)/HOST_PULL;
	    page->mz += (host->z - page->z)/HOST_PULL;
	  }
	  page->x += page->mx; page->y += page->my; page->z += page->mz;
	  page->mx *= SLOWDOWN; page->my *= SLOWDOWN; page->mz *= SLOWDOWN;
	  if ( fabs(page->mx) < MOVE_IS_ACTIVE && 
	       fabs(page->my) < MOVE_IS_ACTIVE &&
	       fabs(page->mz) < MOVE_IS_ACTIVE ) {
	    page->isActive = 0;
	  } else {
	    host->isActive = 1;
	  }
	}
      }
    }
  }
  for ( i=0 ; i<FRAGS_SIZE ; i++ ) {
    if ( frags[i].cnt > 0 ) {
      Frag *fr = &(frags[i]);
      if ( fr->start ) {
	SDL_mutexP(lock);
	fr->txWidth = 0;
	if ( fr->name != NULL && strlen(fr->name) > 0 ) {
	  setTextTexture(fr->name, &(fr->texture),
			 &(fr->txWidth), &(fr->txHeight));
	}
	setFragInitPos(fr, currentPage);
	fr->r = rand()*128/RAND_MAX + 128;
	fr->g = rand()*128/RAND_MAX + 128;
	fr->b = rand()*128/RAND_MAX + 128;
	fr->cnt = rand()*60/RAND_MAX + 60;
	fr->start = 0;
	SDL_mutexV(lock);
      }
      fr->x += fr->mx; fr->y += fr->my; fr->z += fr->mz;
      fr->r *= COLORDOWN; fr->g *= COLORDOWN; fr->b *= COLORDOWN;
      fr->cnt--;
    }
  }
}

#define TITLE_OFFSET 0.25f

static void draw() {
  int i;
  startEnumeration(&hosts);
  while ( hasMoreElements(&hosts) ) {
    Host *host = nextElement(&hosts);
    Hashtable *pages;
    if ( host->start ) continue;
    if ( host->txWidth != 0 ) {
      drawTexture(host->texture,
		  host->x, host->y, host->z,
		  host->txWidth, host->txHeight,
		  100, 255, 255, 0.01f);
    }
    pages = &(host->pages);
    startEnumeration(pages);
    while ( hasMoreElements(pages) ) {
      Page *page = nextElement(pages);
      Hashtable *blinks;
      if ( page->start ) continue;
      drawStar(page->x, page->y, page->z,
	       page->r, page->g, page->b);
      if ( page == currentPage ) {
	drawTwinkleStar(page->x, page->y, page->z,
			92, 92, 92,
			((float)rand()/RAND_MAX)*360, 
			((float)rand()/RAND_MAX)/4+
			((float)rand()/RAND_MAX)/4+0.4f);
      }
      if ( page->txWidth != 0 ) {
	drawTexture(page->texture,
		    page->x, page->y-TITLE_OFFSET, page->z,
		    page->txWidth, page->txHeight,
		    255, 255, 255, 0.005f);
      }
      startDrawLines();
      blinks = &(page->bLinks);
      startEnumeration(blinks);
      while ( hasMoreElements(blinks) ) {
	Link *blink = nextElement(blinks);
	Host *blHost = checkByIndex(&hosts, blink->hostIdx);
	Page *blPage = checkByIndex(&(blHost->pages), blink->pageIdx);
	drawLine(blPage->x, blPage->y, blPage->z,
		 page->x, page->y, page->z,
		 64, 64, 255);
      }
      endDrawLines();
    }
  }
  for ( i=0 ; i<FRAGS_SIZE ; i++ ) {
    if ( frags[i].cnt > 0 ) {
      Frag *fr = &(frags[i]);
      drawStar(fr->x, fr->y, fr->z,
	       fr->r, fr->g, fr->b);
      drawTwinkleStar(fr->x, fr->y, fr->z,
		      fr->g, fr->b, fr->r,
		      ((float)rand()/RAND_MAX)*360, 
		      ((float)rand()/RAND_MAX)/2+
		      ((float)rand()/RAND_MAX)/2+0.5f);
      if ( fr->txWidth != 0 ) {
	drawTexture(fr->texture,
		    fr->x, fr->y-TITLE_OFFSET, fr->z,
		    fr->txWidth, fr->txHeight,
		    255, 255, 255, 0.004f);
      }
    }
  }
}

#define MOUSE_SENSE 0.04f
#define MOUSE_MV_WAIT 10

#define YAW_MV 0.004f
#define YAW_ROLL_MV 0.0025f

static float yawRoll = 0;
static float yawAim = 0, pitchAim = 0; 

static int eyeMove = 0;
static int mouseX, mouseY;
static int mouseMvX, mouseMvY, mousePrsd = 0;
static float eyeSpeed = 0;

static void moveEyePos() {
  float yaOfs;
  if ( eyeMove == 0 ) {
    if ( currentPage != NULL ) {
      eyeX += (currentPage->x - eyeX)/8;
      eyeY += (currentPage->y - eyeY)/8;
      eyeZ += (currentPage->z - eyeZ)/8;
    }
    yawAim += YAW_MV; if ( yawAim >= 6.28f ) yawAim -= 6.28f;
    yawRoll += YAW_ROLL_MV;
    pitchAim = sin(yawRoll)/2;
  } else {
    float mx, my, mz, ms;
    if ( mousePrsd ) {
      yawAim += mouseMvX*MOUSE_SENSE;
      while ( yawAim < 0 ) yawAim += 6.28f;
      while ( yawAim >= 6.28f ) yawAim -= 6.28f;
      pitchAim += mouseMvY*MOUSE_SENSE;
      if ( pitchAim > 1.4 ) pitchAim = 1.4;
      if ( pitchAim < -1.4 ) pitchAim = -1.4;
      mouseMvX = mouseMvY = 0;
      eyeSpeed += 0.001f;
    } else {
      yawAim += YAW_MV/((float)eyeMove/30+1); 
      if ( yawAim >= 6.28f ) yawAim -= 6.28f;
      eyeMove--;
      eyeSpeed *= 0.9f;
    }
    my = -sin(pitch); ms = -cos(pitch);
    mx = -ms * sin(yaw); mz = ms * cos(yaw);
    eyeX += mx*eyeSpeed;
    eyeY += my*eyeSpeed;
    eyeZ += mz*eyeSpeed;
  }
  yaOfs = yawAim - yaw;
  if ( yaOfs < -3.14f ) yaOfs += 6.28f;
  if ( yaOfs >  3.14f ) yaOfs -= 6.28f;
  yaw += yaOfs*0.1f;
  while ( yaw < 0 ) yaw += 6.28f;
  while ( yaw >= 6.28f ) yaw -= 6.28f;
  pitch += (pitchAim-pitch)*0.1f;
}

#define INTERVAL 33
#define AUTO_SAVE_INTERVAL 1000*60*30

static long nextAutoSave;

// Main thread.
int sdlThread(void *data) {
  int done = 0, loaded = 0, i;
  long prvTickCount = 0;
  //int isMinimized = 0;
  SDL_mutex *lock = (SDL_mutex *)data;

  initSDL();
  init();
  nextAutoSave = SDL_GetTicks() + AUTO_SAVE_INTERVAL;

  while ( !done ) {
    long nowTick = SDL_GetTicks();
    int frame = (int)(nowTick-prvTickCount) / INTERVAL;
    if ( frame <= 0 ) {
      frame = 1;
      SDL_Delay(prvTickCount+INTERVAL-nowTick);
      prvTickCount = nowTick;
    } else if ( frame > 5 ) {
      frame = 5;
      prvTickCount = nowTick;
    } else {
      prvTickCount += frame*INTERVAL;
    }
    for ( i=0 ; i<frame ; i++ ) {
      move(lock);
      moveEyePos();
    }
    //if ( !isMinimized ) {
    drawGLSceneStart();
    drawSpace();
    draw();
    drawGLSceneEnd();
    //}

    { SDL_Event event;
      while ( SDL_PollEvent(&event) ) {
	switch ( event.type ) {
	case SDL_QUIT:
          done = 1;
	  break;
	case SDL_VIDEORESIZE:
	  resized(event.resize.w, event.resize.h);
	  break;
	case SDL_MOUSEMOTION:
	  if ( mousePrsd ) {
	    mouseMvX = (event.motion.x-mouseX);
	    mouseMvY = (event.motion.y-mouseY);
	    mouseX = event.motion.x; mouseY = event.motion.y;
	  }
	  break;
	case SDL_MOUSEBUTTONDOWN:
	  if ( event.button.button == SDL_BUTTON_LEFT ) {
	    mousePrsd = 1;
	    mouseX = event.button.x; mouseY = event.button.y;
	    eyeMove = (1000/INTERVAL)*MOUSE_MV_WAIT;
	    eyeSpeed = 0;
	  }
	  break;
	case SDL_MOUSEBUTTONUP:
	  if ( event.button.button == SDL_BUTTON_LEFT ) {
	    mousePrsd = 0;
	  }
	  break;
	}
      }
    }
    if ( !loaded ) {
      loaded = 1;
      loadWindowPref();
    }
    if ( nowTick > nextAutoSave ) {
      nextAutoSave += AUTO_SAVE_INTERVAL;
      saveHistory(lock);
      //printf("Auto saved.\n");
    }
  }
  saveWindowPref();
  saveHistory(lock);
  SDL_Quit();
  exit(1);
}
