/*
 * $Id: screen.c,v 1.5 2001/08/28 01:38:10 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */

/**
 * OpenGL screen handler.
 *
 * @version $Revision: 1.5 $
 */
#include <stdio.h>
#include "kanji.h"
#include "functions.h"
#include "wkf.h"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#if defined(__APPLE__) && defined(__MACH__)
#include <OpenGL/gl.h>	// Header File For The OpenGL32 Library
#include <OpenGL/glu.h>	// Header File For The GLu32 Library
#else
#include <GL/gl.h>	// Header File For The OpenGL32 Library
#include <GL/glu.h>	// Header File For The GLu32 Library
#endif
#include "SDL.h"
#include "SDL_syswm.h"
#include "SDL_ttf.h"

#include <math.h>
#include "sjis2uni.h"

#include "screen.h"

static TTF_Font *font;

#define FAR_PLANE 500

// Init OpenGL with texture, smooth line and alpha.
static void initGL(int width, int height) {
  glViewport(0, 0, width, height);
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glEnable(GL_TEXTURE_2D);

  glEnable(GL_LINE_SMOOTH);
  glLineWidth(2);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, FAR_PLANE);
  glMatrixMode(GL_MODELVIEW);
}

// Reset viewport when the screen is resized.
void resized(int width, int height) {
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, FAR_PLANE);
  glMatrixMode(GL_MODELVIEW);
}

// Load bitmaps and convert to textures.
void loadGLTexture(char *fileName, GLuint *texture) {
  SDL_Surface *surface;

  surface = SDL_LoadBMP(fileName);
  if ( !surface ) {
    fprintf(stderr, "Unable to load texture: %s\n", SDL_GetError());
    SDL_Quit();
    exit(1);
  }

  glGenTextures(1, texture);
  glBindTexture(GL_TEXTURE_2D, *texture);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, surface->w, surface->h, GL_RGB, GL_UNSIGNED_BYTE, surface->pixels);
}

// Create text billboard texture.

static SDL_Surface *createTexture(char *message) {
  SDL_Surface *image;
  SDL_Color fg;
  char sjis_text[BUFSIZ];
  Uint16 unicode_text[BUFSIZ];
  kcode_t kcode;

  kcode = wkfGuessKanjiCodeOfString(message);
  wkfConvertKanjiCodeOfString(kcode, message, KC_SJIS, sjis_text, BUFSIZ);
  sjis2uni(sjis_text, unicode_text);

  fg.r = 0xff; fg.g = 0xff; fg.b = 0xff;
  image = TTF_RenderUNICODE_Blended(font, unicode_text, fg);
  if ( image == NULL ) {
    fprintf(stderr, "Unable to create texture: %s\n", SDL_GetError());
    return(NULL);
  }
  return(image);
}
 
void generateTexture(GLuint *texture) {
  glGenTextures(1, texture);
}

void deleteTexture(GLuint *texture) {
  glDeleteTextures(1, texture);
}

void setTextTexture(char *text, GLuint *texture,
		    GLfloat *width, GLfloat *height) {    
  SDL_Surface *surface;

  surface = createTexture(text);
  if ( !surface ) {
    fprintf(stderr, "Unable to set text texture: %s\n", SDL_GetError());
    *width = 0;
    return;
  }

  glBindTexture(GL_TEXTURE_2D, *texture);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
  *width = surface->w; *height = surface->h;
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, surface->w, surface->h, GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels);
}

void makeTextTexture(char *text, GLuint *texture,
		    GLfloat *width, GLfloat *height) {    
  generateTexture(texture);
  setTextTexture(text, texture, width, height);
}

// Set window start position.
#define WINDOW_POS_FILE "gp.cfg"

#ifdef WIN32
static int loadWindowRect(RECT *rect) {
  FILE *fp = fopen(WINDOW_POS_FILE, "rb");
  if ( fp == NULL ) {
    fprintf(stderr, "Unable to load config file\n");
    return 0;
  }
  rect->left = getw(fp);
  rect->top = getw(fp);
  rect->right = getw(fp);
  rect->bottom = getw(fp);
  fclose(fp);
  return 1;
}

static int saveWindowRect(RECT *rect) {
  FILE *fp = fopen(WINDOW_POS_FILE, "wb");
  if ( fp == NULL ) {
    fprintf(stderr, "Unable to save config file\n");
    return 0;
  }
  putw(rect->left, fp);
  putw(rect->top, fp);
  putw(rect->right, fp);
  putw(rect->bottom, fp);
  fclose(fp);
  return 1;
}
#endif

#define DEFAULT_WINDOW_WIDTH 240
#define DEFAULT_WINDOW_HEIGHT 320

void loadWindowPref() {
#ifdef WIN32
  SDL_SysWMinfo info;
  RECT rect;
  if ( !loadWindowRect(&rect) ) {
    rect.left = rect.top = 0;
    rect.right = DEFAULT_WINDOW_WIDTH;
    rect.bottom = DEFAULT_WINDOW_HEIGHT;
  }
  SDL_GetWMInfo(&info);
  SetWindowPos(info.window, NULL,
	       rect.left, rect.top,
	       rect.right-rect.left, rect.bottom-rect.top,
	       SWP_SHOWWINDOW);
#endif
  SDL_WM_SetCaption("Galactica Proxy", NULL);
}

void saveWindowPref() {
#ifdef WIN32
  SDL_SysWMinfo info;
  RECT rect;
  SDL_GetWMInfo(&info);
  GetWindowRect(info.window, (LPRECT)&rect);
  saveWindowRect(&rect);
#endif
}

static char path[512];

// Get font file name from registry.
static char* getFontFileName() {
#ifdef WIN32
  char file[256];
  HKEY hkey;
  DWORD type = REG_SZ;
  DWORD count = 255;
  if ( RegOpenKeyA(HKEY_CURRENT_USER, 
		   "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",
		   &hkey) ) return NULL;
  if ( RegQueryValueExA(hkey, "Fonts", 0, &type, path, &count)  ) 
    return NULL;
  RegCloseKey(hkey);
  if ( RegOpenKeyA(HKEY_LOCAL_MACHINE,
		   "SYSTEM\\CurrentControlSet\\Control\\FontAssoc\\Associated DefaultFonts",
		   &hkey) ) return NULL;
  if ( RegQueryValueExA(hkey, "AssocSystemFont", 0, &type, file, &count) ) 
    return NULL;
  RegCloseKey(hkey);
  strcat(path, "\\");
  strcat(path, file);
  return path;
#else
  return "Kandata.ttf"
#endif
}

#define SCREEN_WIDTH 640
#define SCREEN_HEIGHT 480

#define PT_SIZE 32

static GLuint starTexture;
#define STAR_BMP "star.bmp"
static GLuint spaceTexture;
#define SPACE_BMP "space.bmp"

void initSDL() {
  char *fontFile;

  /* Initialize SDL for video output */
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
    exit(1);
  }

  /* Initialize the TTF library */
  if ( TTF_Init() < 0 ) {
    fprintf(stderr, "Unable to initialize TTF: %s\n",SDL_GetError());
    exit(2);
  }

  /* Create a OpenGL screen */
  if ( SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 0, 
			SDL_OPENGL | SDL_RESIZABLE ) == NULL ) {
    fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
    SDL_Quit();
    exit(2);
  }

  fontFile = getFontFileName();
  if ( fontFile == NULL ) {
    fprintf(stderr, "Unable to get file info\n");
    exit(2);
  }
  font = TTF_OpenFont(fontFile, PT_SIZE);
  if ( font == NULL ) {
    fprintf(stderr, "Unable to load font from %s: %s\n", 
	    fontFile, SDL_GetError());
    exit(2);
  }
  TTF_SetFontStyle(font, TTF_STYLE_BOLD);

  /* Set the title bar in environments that support it */
  SDL_WM_SetCaption("Now Loading...", NULL);

  initGL(SCREEN_WIDTH, SCREEN_HEIGHT);

  loadGLTexture(STAR_BMP, &starTexture);
  loadGLTexture(SPACE_BMP, &spaceTexture);
}

float eyeX = 0, eyeY = 0, eyeZ = 0;
float yaw = 0;   // 0 < < 6.28f
float pitch = 0; // -1.57f < < 1.57f
float zoom = 10;

static float degrees, degrees2;

// Set eye position.
void drawGLSceneStart() {
  float fDeltaX, fDeltaY, fDeltaZ;

  glClear(GL_COLOR_BUFFER_BIT);

  glPushMatrix();

  fDeltaX = (float)(-sin(yaw) * cos(pitch)) * zoom + eyeX;
  fDeltaY = (float)(sin(pitch)) * zoom + eyeY;
  fDeltaZ = (float)(cos(pitch) * cos(yaw)) * zoom + eyeZ;
    
  gluLookAt(fDeltaX, fDeltaY, fDeltaZ, eyeX, eyeY, eyeZ, 0.0f, 1.0f, 0.0f);

  degrees = (pitch * (180 / 3.1415926535897932385f));
  degrees2 = (yaw * (180 / 3.1415926535897932385f));
}

void drawGLSceneEnd() {
  glPopMatrix();
  SDL_GL_SwapBuffers();
}

void drawStar(GLfloat x, GLfloat y, GLfloat z, int r, int g, int b) {
  glBindTexture(GL_TEXTURE_2D, starTexture);
  glColor4ub(r, g, b, 255);
  glPushMatrix();
  glTranslatef(x, y, z);
  glRotatef(-degrees2, 0.0f, 1.0f, 0.0f);
  glRotatef(-degrees, 1.0f, 0.0f, 0.0f);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  0);
  glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  0);
  glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  0);
  glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  0);
  glEnd();
  glPopMatrix();
}

void drawTwinkleStar(GLfloat x, GLfloat y, GLfloat z, int r, int g, int b,
		     GLfloat d, float size) {
  glBindTexture(GL_TEXTURE_2D, starTexture);
  glColor4ub(r, g, b, 255);
  glPushMatrix();
  glTranslatef(x, y, z);
  glRotatef(-degrees2, 0.0f, 1.0f, 0.0f);
  glRotatef(-degrees, 1.0f, 0.0f, 0.0f);
  glRotatef(d, 0.0f, 0.0f, 1.0f);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0.0f, 0.0f); glVertex3f(-size, -size,  0);
  glTexCoord2f(1.0f, 0.0f); glVertex3f( size, -size,  0);
  glTexCoord2f(1.0f, 1.0f); glVertex3f( size,  size,  0);
  glTexCoord2f(0.0f, 1.0f); glVertex3f(-size,  size,  0);
  glEnd();
  glPopMatrix();
}

void drawTexture(GLuint texture,
		 GLfloat x, GLfloat y, GLfloat z, 
		 GLfloat width, GLfloat height, 
		 int r, int g, int b, float size) {
  glBindTexture(GL_TEXTURE_2D, texture);
  glColor4ub(r, g, b, 255);
  glPushMatrix();
  glTranslatef(x, y, z);
  glRotatef(-degrees2, 0.0f, 1.0f, 0.0f);
  glRotatef(-degrees, 1.0f, 0.0f, 0.0f);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0.0f, 1.0f); 
  glVertex3f(width*-size, height*-size,  0);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f(width*size, height*-size,  0);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f(width*size, height*size,  0);
  glTexCoord2f(0.0f, 0.0f);
  glVertex3f(width*-size, height*size,  0);
  glEnd();
  glPopMatrix();
}

void startDrawLines() {
  glDisable(GL_TEXTURE_2D);
}

void endDrawLines() {
  glEnable(GL_TEXTURE_2D);
}

void drawLine(GLfloat x1, GLfloat y1, GLfloat z1,
	      GLfloat x2, GLfloat y2, GLfloat z2,
	      int r, int g, int b) {
  glColor4ub(r, g, b, 255);
  glBegin(GL_LINES);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y2, z2);
  glEnd();
}

#define SPACE_PLANE 300

void drawSpace() {
  glBindTexture(GL_TEXTURE_2D, spaceTexture);
  glColor4ub(255, 255, 255, 255);
  glPushMatrix();
  glTranslatef(eyeX, eyeY, eyeZ);
  glBegin(GL_QUADS);
  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f( SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);

  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f( SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);

  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f( SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f( SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);

  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);

  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f(-SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f(-SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);

  glTexCoord2f(0.0f, 0.0f); 
  glVertex3f( SPACE_PLANE, -SPACE_PLANE, -SPACE_PLANE);
  glTexCoord2f(0.0f, 1.0f);
  glVertex3f( SPACE_PLANE, -SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 1.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE,  SPACE_PLANE);
  glTexCoord2f(1.0f, 0.0f);
  glVertex3f( SPACE_PLANE,  SPACE_PLANE, -SPACE_PLANE);

  glEnd();
  glPopMatrix();
}
